"""Common functionality of clld Apps is cobbled together here."""
import re
import uuid
import pathlib
import datetime
import functools
import importlib
import collections
import urllib.parse

from sqlalchemy import engine_from_config
from sqlalchemy.orm import joinedload, undefer
from sqlalchemy.exc import NoResultFound

from webob.request import Request as WebobRequest
from zope.interface import implementer, implementedBy
from pyramid.httpexceptions import HTTPNotFound, HTTPMovedPermanently, HTTPGone
from pyramid import events
from pyramid.request import Request, reify
from pyramid.interfaces import IRoutesMapper
from pyramid.asset import abspath_from_asset_spec
from pyramid.renderers import JSON, JSONP
from pyramid.settings import asbool
from clldutils.path import md5, git_describe

import clld
from clld.config import get_config
from clld.db.meta import DBSession, Base
from clld.db.models import common
from clld import Resource, RESOURCES
from clld import interfaces
from clld.web.adapters import get_adapters
from clld.web.adapters import geojson, register_resource_adapters
from clld.web.adapters.base import adapter_factory
from clld.web.views import (
    index_view, resource_view, _raise, _ping, js, unapi, xpartial, redirect, gone,
    select_combination,
)
from clld.web.views.olac import olac, OlacConfig
from clld.web.views.sitemap import robots, sitemapindex, sitemap, resourcemap
from clld.web.subscribers import add_renderer_globals, add_localizer, init_map
from clld.web.datatables.base import DataTable
from clld.web import datatables
from clld.web.maps import Map, ParameterMap, LanguageMap, CombinationMap
from clld.web.icon import ICONS, MapMarker
from clld.web import assets

assert clld
assert assets


class ClldRequest(Request):

    """Custom Request class."""

    @reify
    def query_params(self):
        """Convenient access to the query parameters of the current request.

        :return: dict of the query parameters of the request URL.
        """
        return {k: v[0] for k, v in
                urllib.parse.parse_qs(urllib.parse.urlparse(self.url).query).items()}

    @property
    def db(self):
        """Convenient access to the db session.

        We make the db session available as request attribute, so we do not have to
        import it in templates.
        """
        return DBSession

    @reify
    def dataset(self):
        """Convenient access to the Dataset object.

        Properties of the :py:class:`clld.db.models.common.Dataset` object an
        application serves are used in various places, so we want to have a reference to
        it.
        """
        return self.db.query(common.Dataset).options(undefer('updated')).first()

    @property
    def contact_email_address(self):
        if 'clld.contact' in self.registry.settings:
            return self.registry.settings['clld.contact']
        return self.dataset.contact  # pragma: no cover

    def get_datatable(self, name, model, **kw):
        """Convenient lookup and retrieval of initialized DataTable object.

        :param name: Name under which the datatable class was registered.
        :param model: model class to pass as initialization parameter to the datatable.
        :param kw:
            Keyword parameters are passed through to the initialization of the datatable.
        :return:
            :py:class:`clld.web.datatables.base.DataTable` instance, if a datatable was
            registered for ``name``.
        """
        dt = self.registry.queryUtility(interfaces.IDataTable, name=name)
        if dt:
            return dt(self, model, **kw)

    def get_map(self, name=None, **kw):
        """Convenient lookup and retrieval of initialized Map object.

        :param name: Name under which the map was registered.
        :return:
            :py:class:`clld.web.maps.Map` instance, if a map was registered else ``None``.
        """
        if name is None and self.matched_route:
            name = self.matched_route.name
        if name:
            map_ = self.registry.queryUtility(interfaces.IMap, name=name)
            if map_:
                return map_(self.context, self, **kw)

    def _route(self, obj, rsc, **kw):
        """Determine the name of the canonical route for a resource instance.

        The resource may be specified as object or as mapper class and id.

        :return:
            pair (route_name, kw) suitable as arguments for the Request.route_url method.
        """
        if rsc is None:
            for _rsc in RESOURCES:
                if _rsc.interface.providedBy(obj):
                    rsc = _rsc
                    break
            assert rsc

        route = rsc.name
        if 'ext' in kw:
            route += '_alt'

        # if rsc is passed explicitely, we allow the object id to be passed in as obj,
        # to make it possible to create resource URLs without having the "real" object.
        kw.setdefault('id', getattr(obj, 'id', obj))
        return route, kw

    def ctx_for_url(self, url):
        """Method to reverse URL generation for resources.

        I.e. given a URL, tries to determine the associated resource.

        :return: model instance or ``None``.
        """
        mapper = self.registry.getUtility(IRoutesMapper)
        _path = urllib.parse.urlparse(url).path
        info = mapper(WebobRequest({'PATH_INFO': _path}))
        if not info['route']:
            # FIXME: hack to cater to deployments under a path prefix
            info = mapper(WebobRequest({'PATH_INFO': re.sub(r'^/[a-z]+', '', _path)}))
        if info['route']:
            for rsc in RESOURCES:
                if rsc.name == info['route'].name:
                    if rsc.name == 'dataset':
                        return self.dataset
                    if info['match']:
                        return rsc.model.get(info['match']['id'], default=None)

    def resource_url(self, obj, rsc=None, **kw):
        """Get the absolute URL for a resource.

        :param obj:
            A resource or the id of a resource; in the latter case ``rsc`` must be passed.
        :param rsc: A registered :py:class:`clld.Resource`.
        :param kw:
            Keyword parameters are passed through to
            `pyramid.request.Request.route_url <http://docs.pylonsproject.org/projects/\
            pyramid/en/1.0-branch/api/request.html#pyramid.request.Request.route_url>`_
        :return: URL
        """
        route, kw = self._route(obj, rsc, **kw)
        return self.route_url(route, **kw)

    def route_url(self, route, *args, **kw):
        """Facade for Request.route_url."""
        if '__locale__' in self.params:
            if '_query' not in kw:
                kw['_query'] = {}
            kw['_query']['__locale__'] = self.params['__locale__']
        return Request.route_url(self, route, *args, **kw)

    def resource_path(self, obj, rsc=None, **kw):
        """Determine the path component of a Resource's URL."""
        route, kw = self._route(obj, rsc, **kw)
        return self.route_path(route, **kw)

    def file_ospath(self, file_):
        if 'clld.files' in self.registry.settings:
            return str(self.registry.settings['clld.files'].joinpath(file_.relpath))

    def file_url(self, file_):
        if 'url' in file_.jsondata:
            # just a preparation for full support of non-local files
            return file_.jsondata['url']  # pragma: no cover

        if 'clld.files' in self.registry.settings:
            return self.static_url(self.file_ospath(file_))


def menu_item(route_name, ctx, req, label=None):
    """Factory function for a menu item specified by route name.

    :return: A pair (URL, label) to create a menu item.
    """
    return req.route_url(route_name), label or req.translate(route_name.capitalize())


@implementer(interfaces.ICtxFactoryQuery)
class CtxFactoryQuery(object):

    """Implements reasonable default queries to be used in context factories.

    By reasonable we mean providing good performance for typical data sizes. Applications
    with a-typical numbers of any resource class may have to implement a custom class
    for the ICtxFactoryQuery interface. Usually this will be a class derived from
    CtxFactoryQuery.
    """

    def refined_query(self, query, model, req):
        """To be overridden.

        Derived classes may override this method to add model-specific query
        refinements of their own.
        """
        return query

    def __call__(self, model, req):
        query = req.db.query(model).filter(model.id == req.matchdict['id'])
        custom_query = self.refined_query(query, model, req)

        if query == custom_query:
            # no customizations done, apply the defaults
            f = getattr(model, 'refine_factory_query', None)
            if f:
                query = f(query)
            else:
                if model == common.Contribution:
                    query = query.options(
                        joinedload(
                            common.Contribution.valuesets
                        ).joinedload(
                            common.ValueSet.parameter
                        ),
                        joinedload(
                            common.Contribution.valuesets
                        ).joinedload(
                            common.ValueSet.values
                        ).joinedload(
                            common.Value.domainelement
                        ),
                        joinedload(
                            common.Contribution.references
                        ).joinedload(
                            common.ContributionReference.source
                        ),
                        joinedload(
                            common.Contribution.data
                        )
                    )
        else:
            query = custom_query  # pragma: no cover

        return query.one()


def ctx_factory(model, type_, req):
    """Factory function for request contexts.

    The context of a request is either a single model instance or an instance of
    DataTable incorporating all information to retrieve an appropriately filtered list
    of model instances.
    """
    def replacement(id_):
        raise HTTPMovedPermanently(
            location=req.route_url(model.__name__.lower(), id=id_))

    if type_ == 'index':
        datatable = req.registry.getUtility(
            interfaces.IDataTable, name=req.matched_route.name)
        return datatable(req, model)

    try:
        if model == common.Dataset:
            ctx = req.db.query(model).one()
        elif model == common.Combination:
            ctx = common.Combination.get(req.matchdict['id'])
        else:
            ctx = req.registry.getUtility(interfaces.ICtxFactoryQuery)(model, req)
            if ctx.replacement_id:
                return replacement(ctx.replacement_id)
        ctx.metadata = get_adapters(interfaces.IMetadata, ctx, req)
        return ctx
    except NoResultFound:
        if req.matchdict.get('id'):
            replacement_id = common.Config.get_replacement_id(model, req.matchdict['id'])
            if replacement_id:
                if replacement_id == common.Config.gone:
                    raise HTTPGone()
                return replacement(replacement_id)
        raise HTTPNotFound()


def maybe_import(name, pkg_dir=None):
    exists = False
    if pkg_dir:
        rel_path = name.split('.')[1:] if '.' in name else []
        rel_path.append('__init__.py')
        exists = pkg_dir.joinpath(*rel_path).exists()
        if not exists:
            rel_path.pop()
            if rel_path:
                rel_path[-1] += '.py'
            exists = pkg_dir.joinpath(*rel_path).exists()
    try:
        return importlib.import_module(name)
    except ImportError:
        if pkg_dir and exists:
            print('failed to import existing module {0}'.format(name))
            raise


#
# configurator directives:
#
def register_utility(config, cls, interface, name='', overwrite=True):
    if overwrite or not config.registry.queryUtility(interface, name=name):
        config.registry.registerUtility(cls, provided=interface, name=name)


def register_cls(interface, config, route, cls, overwrite=True):
    register_utility(config, cls, interface, name=route, overwrite=overwrite)
    if not route.endswith('_alt'):
        register_utility(config, cls, interface, name=route + '_alt', overwrite=overwrite)


def register_adapter(config, cls, from_, to_=None, name=None):
    if isinstance(cls, dict):
        cls = adapter_factory(**cls)
    to_ = to_ or list(implementedBy(cls))[0]
    name = name or cls.mimetype
    config.registry.registerAdapter(cls, (from_,), to_, name=name)


def register_adapters(config, specs):
    for interface, base, mimetype, extension, template, extra in specs:
        extra.update(base=base, mimetype=mimetype, extension=extension, template=template)
        config.register_adapter(extra, interface, name=mimetype)


def register_menu(config, *items):
    """Register an item for the main menu.

    :param items: An item may be a (name, factory) pair, where factory is a callable that\
    accepts the two parameters (ctx, req) and returns a pair (url, label) to use for the\
    menu link; or a route name, or a pair (route name, dict), where dict is used as\
    keyword arguments for menu_item.
    """
    menuitems = collections.OrderedDict()
    for item in items:
        if isinstance(item, str):
            item = (item, {})
        name, factory = item
        if isinstance(factory, dict):
            factory = functools.partial(menu_item, name, **factory)
        menuitems[name] = factory
    config.registry.registerUtility(menuitems, interfaces.IMenuItems)


def add_route_and_view(config, route_name, route_pattern, view, **kw):
    """Add a route and a corresponding view and appropriate default routes and views.

    .. note:: To allow custom route patterns we look them up in a dict in settings.
    """
    route_patterns = config.registry.settings.get('route_patterns', {})
    route_pattern = route_patterns.get(route_name, route_pattern)
    alt_route_pattern = kw.pop('alt_route_pattern', route_pattern + '.{ext}')
    route_kw = {}
    factory = kw.pop('factory', None)
    if factory:
        route_kw['factory'] = factory
    config.add_route(route_name, route_pattern, **route_kw)
    config.add_view(view, route_name=route_name, **kw)

    config.add_route(route_name + '_alt', alt_route_pattern, **route_kw)
    config.add_view(view, route_name=route_name + '_alt', **kw)


def register_resource_routes_and_views(config, rsc):
    kw = dict(factory=functools.partial(ctx_factory, rsc.model, 'rsc'))
    if rsc.model == common.Dataset:
        pattern = '/'
        kw['alt_route_pattern'] = '/void.{ext}'
    else:
        pattern = r'/%s/{id:[^/\.]+}' % rsc.plural

    config.add_route_and_view(rsc.name, pattern, resource_view, **kw)
    if rsc.with_index:
        config.add_route_and_view(
            rsc.plural,
            '/%s' % rsc.plural,
            index_view,
            factory=functools.partial(ctx_factory, rsc.model, 'index'))


def register_resource(config, name, model, interface, with_index=False, **kw):
    """Directive to register custom resources.

    .. note::

        The directive accepts arbitrary keyword arguments for backwards compatibility.
    """
    # in case of tests, this method may be called multiple times!
    if [rsc for rsc in RESOURCES if rsc.name == name]:
        return
    rsc = Resource(name, model, interface, with_index=with_index)
    RESOURCES.append(rsc)
    config.register_resource_routes_and_views(rsc)

    if not config.registry.queryUtility(interfaces.IDataTable, name=rsc.plural):
        config.register_datatable(
            rsc.plural, getattr(datatables, rsc.plural.capitalize(), DataTable))

    register_resource_adapters(config, rsc)


def register_download(config, download):
    config.registry.registerUtility(download, interfaces.IDownload, name=download.name)


StaticResource = collections.namedtuple('StaticResource', 'type asset_spec')


def register_staticresource(config, type, asset_spec):
    config.registry.registerUtility(
        StaticResource(type, asset_spec), interfaces.IStaticResource, name=asset_spec)


def add_settings_from_file(config, file_):
    if file_.exists():
        cfg = get_config(file_)
        if 'mako.directories_list' in cfg:
            cfg['mako.directories'] = cfg['mako.directories_list']  # pragma: no cover
        config.add_settings(cfg)


def _route_and_view(config, pattern, view, name=None):
    name = name or str(uuid.uuid4())
    config.add_route(name, pattern)
    config.add_view(view, route_name=name)


def add_301(config, pattern, location, name=None):
    _route_and_view(
        config, pattern, xpartial(redirect, HTTPMovedPermanently, location), name=name)


def add_410(config, pattern, name=None):
    _route_and_view(config, pattern, gone, name=name)


def add_page(config, name, pattern=None, view=None, template=None, views=None):
    views = views or maybe_import('%s.views' % config.root_package.__name__)
    config.add_route_and_view(
        name,
        pattern or '/' + name,
        view or getattr(views, name, lambda r: {}),
        renderer=template or name + '.mako')


def includeme(config):
    """Upgrading:

    - register utilities "by hand", after config.include('clld.web.app')
    - add routes by hand (and remove these from the **kw passed to Configurator)

    :param config:
    :return:
    """
    #
    # now we exploit the default package layout as created via the CLLD scaffold:
    #
    # note: the following exploits the import time side effect of modifying the webassets
    # environment!
    root_package = config.root_package.__name__
    pkg_dir = pathlib.Path(config.root_package.__file__).parent.resolve()
    maybe_import('%s.assets' % root_package, pkg_dir=pkg_dir)

    json_renderer = JSON()
    json_renderer.add_adapter(datetime.datetime, lambda obj, req: obj.isoformat())
    json_renderer.add_adapter(datetime.date, lambda obj, req: obj.isoformat())
    config.add_renderer('json', json_renderer)

    jsonp_renderer = JSONP(param_name='callback')
    jsonp_renderer.add_adapter(datetime.datetime, lambda obj, req: obj.isoformat())
    jsonp_renderer.add_adapter(datetime.date, lambda obj, req: obj.isoformat())
    config.add_renderer('jsonp', jsonp_renderer)

    config.set_request_factory(ClldRequest)
    config.registry.registerUtility(CtxFactoryQuery(), interfaces.ICtxFactoryQuery)
    config.registry.registerUtility(OlacConfig(), interfaces.IOlacConfig)

    # initialize the db connection
    engine = engine_from_config(config.registry.settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine

    try:
        git_tag = git_describe(pathlib.Path(pkg_dir).parent)
    except ValueError:  # pragma: no cover
        git_tag = None

    config.add_settings({
        'pyramid.default_locale_name': 'en',
        'clld.pkg': root_package,
        'clld.git_tag': git_tag,
        'clld.parameters': {}})
    if 'clld.files' in config.registry.settings:
        # deployment-specific location of static data files
        abspath = pathlib.Path(config.registry.settings['clld.files']).resolve()
        config.add_settings({'clld.files': abspath})
        config.add_static_view('files', str(abspath))

    # event subscribers:
    config.add_subscriber(add_localizer, events.NewRequest)
    config.add_subscriber(init_map, events.ContextFound)
    config.add_subscriber(
        functools.partial(
            add_renderer_globals,
            maybe_import('%s.util' % root_package, pkg_dir=pkg_dir)),
        events.BeforeRender)

    #
    # make it easy to register custom functionality
    #
    for name, func in {
        'register_utility': register_utility,
        'register_datatable': functools.partial(register_cls, interfaces.IDataTable),
        'register_map': functools.partial(register_cls, interfaces.IMap),
        'register_menu': register_menu,
        'register_resource': register_resource,
        'register_adapter': register_adapter,
        'register_adapters': register_adapters,
        'register_download': register_download,
        'register_staticresource': register_staticresource,
        'add_route_and_view': add_route_and_view,
        'add_settings_from_file': add_settings_from_file,
        'add_301': add_301,
        'add_410': add_410,
        'add_page': add_page,
        'register_resource_routes_and_views': register_resource_routes_and_views,
    }.items():
        config.add_directive(name, func)

    #
    # routes and views
    #
    config.add_static_view('clld-static', 'clld:web/static')
    config.add_static_view('static', '%s:static' % root_package)

    config.add_route_and_view('_js', '/_js', js, http_cache=3600)

    # add some maintenance hatches
    config.add_route_and_view('_raise', '/_raise', _raise)
    config.add_route_and_view('_ping', '/_ping', _ping, renderer='json')

    # sitemap support:
    config.add_route_and_view('robots', '/robots.txt', robots)
    config.add_route_and_view('sitemapindex', '/sitemap.xml', sitemapindex)
    config.add_route_and_view('sitemap', '/sitemap.{rsc}.{n}.xml', sitemap)
    config.add_route('resourcemap', '/resourcemap.json')
    config.add_view(resourcemap, route_name='resourcemap', renderer='jsonp')
    config.add_route_and_view(
        'select_combination', '/_select_combination', select_combination)

    config.add_route_and_view('unapi', '/unapi', unapi)
    config.add_route_and_view('olac', '/olac', olac)

    config.add_settings_from_file(pkg_dir.joinpath('appconf.ini'))
    if not config.registry.settings.get('mako.directories'):
        config.add_settings({'mako.directories': ['clld:web/templates']})

    for rsc in RESOURCES:
        config.register_resource_routes_and_views(rsc)
        config.register_datatable(
            rsc.plural, getattr(datatables, rsc.plural.capitalize(), DataTable))
        register_resource_adapters(config, rsc)

    # maps
    config.register_map('languages', Map)
    config.register_map('language', LanguageMap)
    config.register_map('parameter', ParameterMap)
    config.register_map('combination', CombinationMap)

    config.include('clld.web.adapters')

    for icon in ICONS:
        config.registry.registerUtility(icon, interfaces.IIcon, name=icon.name)
    config.registry.registerUtility(MapMarker(), interfaces.IMapMarker)

    #
    # inspect default locations for views and templates:
    #
    home_comp = collections.OrderedDict()
    for name, template in [
        ('introduction', False),
        ('about', False),
        ('terms', False),
        ('glossary', False),
        ('history', False),
        ('changes', False),
        ('credits', False),
        ('legal', True),
        ('download', True),
        ('contact', True),
        ('help', False),
    ]:
        home_comp[name] = template

    if pkg_dir.joinpath('templates').exists():
        for p in pkg_dir.joinpath('templates').iterdir():
            if p.stem in home_comp and p.suffix == '.mako':
                home_comp[p.stem] = True

    for name, template in home_comp.items():
        if template:
            config.add_page(name)

    config.add_settings({'home_comp': [k for k in home_comp.keys() if home_comp[k]]})

    if 'clld.favicon' not in config.registry.settings:
        favicon = {'clld.favicon': 'clld:web/static/images/favicon.ico'}
        # hard to test (in particular on travis) and without too much consequence
        # (and the consequences faced are easy to spot).
        if pkg_dir.joinpath('static', 'favicon.ico').exists():  # pragma: no cover
            favicon['clld.favicon'] = root_package + ':static/favicon.ico'
        config.add_settings(favicon)

    config.add_settings({
        'clld.favicon_hash': md5(abspath_from_asset_spec(
            config.registry.settings['clld.favicon']))})

    translation_dirs = ['clld:locale']
    if pkg_dir.joinpath('locale').exists():
        translation_dirs.append('%s:locale' % root_package)  # pragma: no cover
    config.add_translation_dirs(*translation_dirs)

    if pkg_dir.joinpath('static/publisher_logo.png').exists():  # pragma: no cover
        config.add_settings(
            {'clld.publisher_logo': '%s:static/publisher_logo.png' % root_package})

    if asbool(config.registry.settings.get('clld.pacific_centered_maps')):
        geojson.pacific_centered()

    v = maybe_import('%s.views' % root_package, pkg_dir=pkg_dir)
    if v:
        config.scan(v)  # pragma: no cover

    menuitems = config.registry.settings.get(
        'clld.menuitems_list',
        ['contributions', 'parameters', 'languages', 'contributors'])
    config.register_menu(
        ('dataset', lambda ctx, req: (req.resource_url(req.dataset), req.translate('Home'))),
        *menuitems)

    config.include('pyramid_mako')

    for name in ['adapters', 'datatables', 'maps']:
        mod = maybe_import('%s.%s' % (root_package, name), pkg_dir=pkg_dir)
        if mod and hasattr(mod, 'includeme'):
            config.include(mod)



"""Implementation of a component concept."""
import functools

from markupsafe import Markup
from pyramid.renderers import render


class Component(object):

    """Virtual base class for page components.

    Components are objects that can be rendered as HTML
    and typically define behavior using a corresponding JavaScript object which accepts
    an options object upon initialization.
    """

    __template__ = None

    @functools.cached_property
    def options(self):
        """Typically options to configure a corresponding JavaScript object.

        :return: JSON serializable dict
        """
        opts = self.get_default_options()
        opts.update(self.get_options() or {})
        opts.update(self.get_options_from_req() or {})
        return opts

    def render(self):
        return Markup(render(
            self.__template__, {'obj': self}, request=getattr(self, 'req', None)))

    def get_options(self):
        """Override this method to define final-class-specific options.

        :return: JSON serializable dict
        """
        return {}

    def get_default_options(self):
        """Override this method to define default (i.e. valid across subclasses) options.

        :return: JSON serializable dict
        """
        return {}  # pragma: no cover

    def get_options_from_req(self):
        """Override this method to define options derived from request properties.

        :return: JSON serializable dict
        """
        return {}



"""Default DataTable for Contribution objects."""
from clld.web.datatables.base import DataTable, Col, LinkCol
from clld.web.util.helpers import linked_contributors, cite_button


class ContributorsCol(Col):

    """Render links to the corresponding Contributors of a Contribution."""

    __kw__ = {'bSearchable': False, 'bSortable': False}

    def format(self, item):
        return linked_contributors(self.dt.req, item)


class CitationCol(Col):

    """Render a cite-button."""

    __kw__ = {'bSearchable': False, 'bSortable': False}

    def format(self, item):
        return cite_button(self.dt.req, item)


class Contributions(DataTable):

    """Default DataTable for Contribution objects."""

    def col_defs(self):
        return [
            LinkCol(self, 'name'),
            ContributorsCol(self, 'contributor'),
            CitationCol(self, 'cite'),
        ]



"""Base classes for adapters."""
import uuid

from zope.interface import implementer
from pyramid.response import Response
from pyramid.renderers import render as pyramid_render
from clldutils.misc import slug

from clld import interfaces


class Renderable:

    """Virtual base class for adapters.

    Adapters can provide custom behaviour either by specifying a template to use for
    rendering, or by overwriting the render method.
    """

    name = None
    template = None
    mimetype = 'text/plain'
    extension = None
    send_mimetype = None
    rel = 'alternate'

    def __init__(self, obj):
        self.obj = obj

    @property
    def label(self):
        return getattr(self, '__label__', self.__class__.__name__)

    @property
    def charset(self):
        return 'utf-8' \
            if self.mimetype.startswith('text/') \
            or 'xml' in self.mimetype \
            or 'kml' in self.mimetype \
            else None

    def render_to_response(self, ctx, req):
        res = Response(self.render(ctx, req))
        res.vary = str('Accept')
        res.content_type = str(self.send_mimetype or self.mimetype)
        if self.charset:
            res.content_type += str('; charset=') + str(self.charset)
        return res

    def template_context(self, ctx, req):
        return {}

    def render(self, ctx, req):
        context = self.template_context(ctx, req)
        context.setdefault('ctx', ctx)
        return pyramid_render(self.template, context, request=req)


@implementer(interfaces.IRepresentation)
class Representation(Renderable):

    """Base class for adapters implementing IRepresentation."""


@implementer(interfaces.IRepresentation)
class Json(Renderable):

    """JavaScript Object Notation."""

    name = 'JSON'
    mimetype = 'application/json'
    extension = 'json'

    def render(self, ctx, req):
        return pyramid_render('json', ctx, request=req)


@implementer(interfaces.IIndex)
class Index(Renderable):

    """Base class for adapters implementing IIndex."""


def adapter_factory(*args, **kw):
    # for backwards compatibility we interpret the first positional argument as template:
    if args:
        kw['template'] = args[0]
    assert 'template' in kw
    kw.setdefault('mimetype', 'text/html')
    kw.setdefault('extension', 'html')
    base = kw.pop('base', Representation)
    return type(str('AdapterFromFactory%s' % slug(str(uuid.uuid4()))), (base,), kw)



"""Functionality to configure leaflet maps from python."""
import functools

from clld.interfaces import IDataTable, IMapMarker, IIcon
from clld.web.util import helpers
from clld.web.util.htmllib import HTML
from clld.web.util.component import Component
from clld.web.adapters.geojson import GeoJson, GeoJsonCombinationDomainElement, get_lonlat


class Layer(object):

    """Represents a layer in a leaflet map.

    A layer in our terminology is a
    `FeatureCollection <http://geojson.org/geojson-spec.html#feature-collection-objects>`_
    in geojson and a
    `geoJson layer <http://leafletjs.com/reference.html#geojson>`_
    in leaflet, i.e. a bunch of points on the map.
    """

    def __init__(self, id_, name, data, **kw):
        """Initialize a layer object.

        :param id_: Map-wide unique string identifying the layer.
        :param name: Human readable name of the layer.
        :param data: A GeoJSON FeatureCollection either specified as corresponding Python\
        dict or as URL which will serve the appropriate GeoJSON.
        :param kw: Additional keyword parameters are made available to the Layer as \
        instance attributes.
        """
        self.id = id_
        self.name = name
        self.data = data
        self.addToLayersControl = False
        for k, v in kw.items():
            setattr(self, k, v)


class Legend(object):

    """Represents a navpill with a dropdown above a map."""

    def __init__(self,
                 map_,
                 name,
                 items,
                 label=None,
                 stay_open=False,
                 item_attrs=None,
                 pull_right=False):
        self.map = map_
        self.name = name
        self.label = label or name.capitalize()
        self.items = items
        self.stay_open = stay_open
        self.item_attrs = item_attrs or {}
        self.pull_right = pull_right

    def format_id(self, suffix=None):
        suffix = suffix or ''
        if suffix:
            suffix = '-' + suffix
        return 'legend-%s%s' % (self.name, suffix)

    def render_item(self, item):
        if not isinstance(item, (tuple, list)):
            item = [item]
        attrs = self.item_attrs
        if self.stay_open:
            class_ = attrs.get('class', attrs.get('class_', ''))
            attrs['class'] = class_ + ' stay-open'
        return HTML.li(*item, **attrs)

    def render(self):
        a_attrs = {
            'class': 'dropdown-toggle',
            'data-toggle': "dropdown",
            'href': "#",
            'id': self.format_id('opener')}
        ul_class = 'dropdown-menu'
        if self.stay_open:
            ul_class += ' stay-open'
        return HTML.li(
            HTML.a(self.label, HTML.b(class_='caret'), **a_attrs),
            HTML.ul(
                *map(self.render_item, self.items),
                **dict(class_=ul_class, id=self.format_id('container'))),
            class_='dropdown' + (' pull-right' if self.pull_right else ''),
            id=self.format_id(),
        )


class FilterLegend(Legend):

    """Legend with actionable items.

    Legend rendering radio controls to filter languages on a map in sync with a column
    of an associated DataTable.
    """

    def __init__(self, map_, value_getter, col=None, dt=None, **kw):
        """Initialize.

        @param value_getter: Name of a javascript object which will be called with the \
        properties associated with a map marker to determine its filter value.
        """
        kw.setdefault('stay_open', True)
        if col and dt:
            col = {c.name: c for c in dt.cols}.get(col)
        if col:
            kw.setdefault('label', col.js_args.get('sTitle'))
        Legend.__init__(self, map_, col.name if col else 'nocol', [], **kw)
        self.jsname = 'fl-' + self.name
        items = [self.li(col, '--any--', value_getter, checked=True)]
        for item in getattr(col, 'choices', []):
            items.append(self.li(col, item, value_getter))
        self.items = items

    def li_label(self, item):
        return item[1] if isinstance(item, (tuple, list)) else item

    def li(self, col, item, value_getter, checked=False):
        input_attrs = dict(
            type='radio',
            class_='stay-open %s inline' % self.jsname,
            name=self.jsname,
            value=item[0] if isinstance(item, (tuple, list)) else item,
            onclick=helpers.JS("CLLD.mapLegendFilter")(
                self.map.eid,
                self.name,
                self.jsname,
                helpers.JS(value_getter),
                col.dt.eid if col else None))
        if checked:
            input_attrs['checked'] = 'checked'
        return HTML.label(
            HTML.input(**input_attrs),
            ' ',
            self.li_label(item),
            class_="stay-open",
            style="margin-left:5px; margin-right:5px;",
        )


class Map(Component):
    """
    Represents the configuration for a leaflet map.

    Map options to be specified as items returned in `get_options`:

    - info_route: Name of the route to request info window content ['language_alt']
    - info_query: Query parameters to pass when requesting info window content [{}]
    - overlays: list of dicts(name=, url=, options=), specifying tileOverlays [[]]
    - exclude_from_zoom: list of layer names to exclude when zooming to extent [[]]
    - base_layer: Name of base layer to use.
    - no_popup: Flag to opt out of info window popups [False]
    - no_link: Flag to opt out of turning markers into links [False]
    - icons: Name of icons style as defined in `CLLD.MapIcons` ['base']
    - icon_size: initial size of map markers in pixels [20]
    - sidebar: Flag indicating whether the map is rendered in the sidebar [False]
    - zoom: Integer specifying the initial zoom factor of the map [None]
    - max_zoom: Integer specfying the maximal zoom factor allowed for a map [6]
    - center: (latitutde, longitude) pair specifying the center of the map [None]
    - hash: Flag indicating whether map center coords should be added to the URL [False]
    - tile_layer: dict(url_pattern=, options=), specifying an alternative base layer [None]
    - add_layers_to_control: Flag indicating whether map layers should be added to the \
      layers control as overlays [False]
    - show_labels: Flag indicating whether labels/tooltips should be initially open [False]
    - on_init: function to run at the end of map initialization [None]
    - resize_direction: 'e', 's' or 'se', make map resizeable.
      see https://github.com/jjimenezshaw/Leaflet.Control.Resizer#api
    - with_audioplayer: Flag indicating whether to add an AudioPlayer control on the map.

    """

    __template__ = 'clld:web/templates/map.mako'

    def __init__(self, ctx, req, eid='map'):
        """Initialize.

        :param ctx: context object of the current request.
        :param req: current pyramid request object.
        :param eid: Page-unique DOM-node ID.
        """
        self.req = req
        self.ctx = ctx
        self.eid = eid
        self.map_marker = req.registry.getUtility(IMapMarker)

    def get_options_from_req(self):
        params = self.req.params
        res = {}
        try:
            if 'lat' in params and 'lng' in params:
                res['center'] = list(map(float, [params['lat'], params['lng']]))
            if 'z' in params:
                res['zoom'] = int(params['z'])
            if 'iconsize' in params:
                res['icon_size'] = int(params['iconsize'])
            if 'labels' in params:
                res['show_labels'] = params['labels'] == '1'
        except (ValueError, TypeError):
            pass
        return res

    def get_default_options(self):
        return {'resize_direction': 's'}

    @functools.cached_property
    def layers(self):
        """The list of layers of the map.

        .. note:: Since layers may be costly to compute, we cache them per map instance.

        :return: list of :py:class:`clld.web.maps.Layer` instances.
        """
        return list(self.get_layers())

    def get_layers(self):
        """Generate the list of layers.

        :return: list or generator of :py:class:`clld.web.maps.Layer` instances.
        """
        route_params = {'ext': 'geojson'}
        if not IDataTable.providedBy(self.ctx):
            route_params['id'] = self.ctx.id
        route_name = self.req.matched_route.name
        if not route_name.endswith('_alt'):
            route_name += '_alt'
        yield Layer(
            getattr(self.ctx, 'id', 'id'),
            '%s' % self.ctx,
            self.req.route_url(route_name, **route_params))

    @functools.cached_property
    def legends(self):
        return list(self.get_legends())

    def get_legends(self):
        if len(self.layers) > 1:
            items = []
            total = 0
            repr_attrs = dict(class_='pull-right stay-open', style="padding-right: 10px;")

            for layer in self.layers:
                representation = ''
                if hasattr(layer, 'representation'):
                    total += layer.representation
                    representation = HTML.span(str(layer.representation), **repr_attrs)
                items.append([
                    HTML.label(
                        HTML.input(
                            class_="stay-open",
                            type="checkbox",
                            checked="checked",
                            onclick=helpers.JS_CLLD.mapToggleLayer(
                                self.eid, layer.id, helpers.JS("this"))),
                        getattr(layer, 'marker', ''),
                        layer.name,
                        class_="checkbox inline stay-open",
                        style="margin-left: 5px; margin-right: 5px;",
                    ),
                    representation,
                ])
            if total:
                items.append(HTML.span(HTML.b(str(total)), **repr_attrs))
            yield Legend(
                self,
                'layers',
                items,
                label='Legend',
                stay_open=True,
                item_attrs=dict(style='clear: right'))
        items = []
        iconsize = self.options.get('icon_size', 30)
        for size in [15, 20, 30, 40]:
            attrs = dict(name="iconsize", value=str(size), type="radio")
            if size == iconsize:
                attrs['checked'] = 'checked'
            items.append(HTML.label(
                HTML.input(onclick=helpers.JS_CLLD.mapResizeIcons(self.eid), **attrs),
                HTML.img(
                    height=str(size),
                    width=str(size),
                    src=self.req.registry.getUtility(IIcon, 'cff6600').url(self.req)),
                class_="radio",
                style="margin-left: 5px; margin-right: 5px;"))
        yield Legend(
            self,
            'iconsize',
            items,
            label=self.req.translate('Icon size'))

        def item(layer):
            return HTML.a(
                layer.name,
                onclick='return %s;' % helpers.JS_CLLD.mapShowGeojson(self.eid, layer.id),
                href=layer.data if isinstance(layer.data, str) else '#')
        yield Legend(
            self, 'geojson', map(item, self.layers), label='GeoJSON', pull_right=True)


class ParameterMap(Map):

    """Map displaying markers for valuesets associated with a parameter instance."""

    def get_layers(self):
        if self.ctx.domain:
            for de in self.ctx.domain:
                yield Layer(
                    de.id,
                    de.name,
                    self.req.resource_url(
                        self.ctx, ext='geojson',
                        _query=dict(domainelement=str(de.id), **self.req.query_params)
                    ),
                    marker=helpers.map_marker_img(self.req, de, marker=self.map_marker))
        else:
            yield Layer(
                self.ctx.id,
                self.ctx.name,
                self.req.resource_url(self.ctx, ext='geojson'))

    def get_default_options(self):
        return {
            'resize_direction': 's',
            'info_query': {'parameter': self.ctx.pk},
            'hash': True}


class GeoJsonMultiple(GeoJson):

    """Render a collection of languages as geojson feature collection."""

    def feature_iterator(self, ctx, req):
        return ctx

    def feature_properties(self, ctx, req, language):
        return {
            'icon': req.registry.getUtility(IIcon, 'tff0000').url(req),
            'icon_size': 10,
            'zindex': 1000}


class CombinationMap(Map):

    """Map for a combination of parameters."""

    def get_layers(self):
        for de in self.ctx.domain:
            if de.languages:
                yield Layer(
                    de.id,
                    de.name,
                    GeoJsonCombinationDomainElement(de).render(de, self.req, dump=False),
                    marker=HTML.img(src=de.icon.url(self.req), height='20', width='20'))
        if self.ctx.multiple:
            # yield another layer which can be used to mark languages with multiple
            # values, because this may not be visible when markers are stacked on top
            # of each other.
            icon_url = self.req.registry.getUtility(IIcon, 'tff0000').url(self.req)
            yield Layer(
                '__multiple__',
                'Languages with multiple values',
                GeoJsonMultiple(None).render(self.ctx.multiple, self.req, dump=False),
                marker=HTML.img(src=icon_url, height='20', width='20'))

    def get_options(self):
        return {'icon_size': 25, 'hash': True}


class LanguageMap(Map):

    """Map showing a single language."""

    def get_layers(self):
        yield Layer(
            self.ctx.id,
            self.ctx.name,
            GeoJson(self.ctx).render(self.ctx, self.req, dump=False))

    def get_default_options(self):
        return {
            'center': list(reversed(get_lonlat(self.ctx) or [0, 0])),
            'zoom': 3,
            'no_popup': True,
            'no_link': True,
            'sidebar': True}


class GeoJsonSelectedLanguages(GeoJson):

    """Represents the geo-data of an iterable selection of languages.

    The iterable is assumed to be passed into the adapter upon initialization.
    """

    def feature_iterator(self, ctx, req):
        return self.obj


class SelectedLanguagesMap(Map):

    """Map showing an arbitrary selection of languages."""

    def __init__(self, ctx, req, languages, geojson_impl=None, **kw):
        """Initialize.

        :param languages: Iterable collection of languages.
        :param geojson_impl: GeoJson implementation to use.
        """
        self.geojson_impl = geojson_impl or GeoJsonSelectedLanguages
        self.languages = languages
        Map.__init__(self, ctx, req, **kw)

    def get_options(self):
        return {'icon_size': 20, 'hash': True, 'show_labels': len(self.languages) < 100}

    def get_layers(self):
        yield Layer(
            'languages',
            'Languages',
            self.geojson_impl(self.languages).render(self.ctx, self.req, dump=False))



import pytest

from clld.web.util.glottolog import *


def test_url():
    assert url().endswith('glottolog.org')
    assert url('1').endswith('languoid/id/1')
    assert url('x', obj_type='reference').endswith('reference/id/x')

    with pytest.raises(AssertionError):
        url('x', obj_type='x')


def test_link(env):
    assert '.png' in link(env['request'], id='1')



from clld.db.models import common

from testutils import handle_dt


def test_Parameters(env):
    from clld.web.datatables.parameter import Parameters

    handle_dt(env['request'], Parameters, common.Parameter)



import pytest
from sqlalchemy.exc import NoResultFound

from clld.db.models.common import Language
from clld.db.meta import DBSession, is_base


def test_JSONEncodedDict(db, persist):
    l = persist(Language(id='abc', name='Name', jsondata={'i': 2}))
    DBSession.expunge(l)
    for lang in DBSession.query(Language).filter(Language.id == 'abc'):
        assert lang.jsondata['i'] == 2
        break


def test_CustomModelMixin(db, custom_language, persist):
    persist(custom_language(id='abc', name='Name', custom='c'))
    for lang in DBSession.query(Language).filter(Language.id == 'abc'):
        assert lang.custom == 'c'
        break


def test_CustomModelMixin_polymorphic(db, custom_language, persist):
    lang = Language(id='def', name='Name')
    assert repr(lang).startswith("<Language ")
    assert is_base(Language)
    assert not is_base(custom_language)
    clang = custom_language(id='abc', name='Name', custom='c')
    persist(lang, clang)
    DBSession.expunge_all()
    lang = DBSession.query(Language).filter_by(id='def').one()
    clang = DBSession.query(Language).filter_by(id='abc').one()

    assert lang.polymorphic_type == 'base'
    assert clang.polymorphic_type == 'custom'
    assert type(lang) is Language
    assert type(clang) is custom_language


def test_Base(db, persist):
    l = persist(Language(id='abc', name='Name'))
    DBSession.expunge(l)
    l = Language.get('abc', session=DBSession)
    assert l.name == 'Name'

    Language().__str__()
    assert repr(l) == "<Language 'abc'>"


def test_Base_jsondata(db, persist):
    l = persist(Language(id='abc', name='Name'))
    l.update_jsondata(a=1)
    assert 'a' in l.jsondata
    l.update_jsondata(b=1)
    assert 'b' in l.jsondata and 'a' in l.jsondata
    assert 'b' in l.__json__(None)['jsondata']


def test_Base_get(db):
    assert 42 == Language.get('doesntexist', default=42)
    with pytest.raises(NoResultFound):
        Language.get('doesntexist')



from sqlalchemy import (
    Column,
    Float,
    Integer,
    Unicode,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    Sentence, ValueSet,
    parameter as _parameter)  # needed to initalize relationship
assert _parameter

__all__ = ('Value', 'ValueSentence')


class Value_data(Base, DataMixin):
    pass


class Value_files(Base, FilesMixin):
    pass


@implementer(interfaces.IValue)
class Value(Base,
            PolymorphicBaseMixin,
            IdNameDescriptionMixin,
            HasDataMixin,
            HasFilesMixin):

    """A measurement of a parameter for a particular language."""

    __table_args__ = (
        UniqueConstraint('valueset_pk', 'name', 'domainelement_pk'),
    )

    # we must override the pk col declaration from Base to have it available for ordering.
    pk = Column(Integer, primary_key=True)
    valueset_pk = Column(Integer, ForeignKey('valueset.pk'), nullable=False)
    # Values may be taken from a domain.
    domainelement_pk = Column(Integer, ForeignKey('domainelement.pk'))

    frequency = Column(
        Float,
        doc='frequency of the value relative to other values for the same language')
    """Languages may have multiple values for the same parameter. Their relative
    frequency can be stored here."""
    confidence = Column(
        Unicode, doc='textual assessment of the reliability of the value assignment')

    domainelement = relationship('DomainElement', backref='values')

    @declared_attr
    def valueset(cls):
        return relationship(
            ValueSet, innerjoin=True,
            backref=backref(
                'values', order_by=[cls.frequency.desc(), cls.confidence, cls.pk]))

    def __json__(self, req):
        res = Base.__json__(self, req)
        res['domainelement'] = self.domainelement.__json__(req) \
            if self.domainelement else None
        res['valueset'] = self.valueset.__json__(req)
        return res

    def __str__(self):
        if self.domainelement and self.domainelement.name:
            return self.domainelement.name
        else:
            return self.name or self.id


class ValueSentence(Base, PolymorphicBaseMixin):

    """Association between values and sentences given as explanation of a value."""

    __table_args__ = (UniqueConstraint('value_pk', 'sentence_pk'),)

    value_pk = Column(Integer, ForeignKey('value.pk'), nullable=False)
    sentence_pk = Column(Integer, ForeignKey('sentence.pk'), nullable=False)
    description = Column(Unicode())

    value = relationship(Value, innerjoin=True, backref='sentence_assocs')
    sentence = relationship(Sentence, innerjoin=True, backref='value_assocs', order_by=Sentence.id)



import collections

from pyramid.config import Configurator
from clld.interfaces import IMapMarker, IValueSet, IValue, IDomainElement
from clld.web.icon import MapMarker
from clldutils.svg import pie, icon, data_url

# we must make sure custom models are known at database initialization!
from {{cookiecutter.directory_name}} import models


{% if cookiecutter.cldf_module.lower() == 'structuredataset' %}
class FeatureMapMarker(MapMarker):
    def __call__(self, ctx, req):
        if IValueSet.providedBy(ctx):
            c = collections.Counter([v.domainelement.jsondata['color'] for v in ctx.values])
            return data_url(pie(*list(zip(*[(v, k) for k, v in c.most_common()])), **dict(stroke_circle=True)))
        if IDomainElement.providedBy(ctx):
            return data_url(icon(ctx.jsondata['color'].replace('#', 'c')))
        if IValue.providedBy(ctx):
            return data_url(icon(ctx.domainelement.jsondata['color'].replace('#', 'c')))
        return super(FeatureMapMarker, self).__call__(ctx, req)
{% endif %}


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    config = Configurator(settings=settings)
    config.include('clld.web.app')
{% if cookiecutter.mpg == true %}
    config.include('clldmpg')
{% endif %}
{% if cookiecutter.cldf_module.lower() == 'structuredataset' %}
    config.registry.registerUtility(FeatureMapMarker(), IMapMarker)
{% endif %}
    return config.make_wsgi_app()



"""Default DataTable for Sentence objects."""
from sqlalchemy import and_
from sqlalchemy.orm import joinedload

from clld.db.util import get_distinct_values
from clld.db.models.common import (
    Language, Sentence, Parameter, ValueSentence, Value, ValueSet, Sentence_files,
)
from clld.web.datatables.base import (
    DataTable, LinkCol, DetailsRowLinkCol, Col,
)
from clld.web.util.helpers import icon


class TypeCol(Col):

    """Render the type attribute of a Sentence."""

    def __init__(self, dt, name, **kw):
        kw.setdefault('sTitle', dt.req.translate('Type'))
        kw.setdefault('choices', get_distinct_values(Sentence.type))
        super(TypeCol, self).__init__(dt, name, **kw)

    def search(self, qs):
        return Sentence.type == qs

    def format(self, item):
        return item.type

    def order(self):
        return Sentence.type


class TsvCol(Col):
    def search(self, qs):
        return super(TsvCol, self).search('\t'.join(qs.split()))


class AudioCol(Col):
    def __init__(self, dt, name, **kw):
        kw['choices'] = ['yes']
        kw['input-size'] = 'mini'
        kw['model_col'] = Sentence_files.id
        Col.__init__(self, dt, name, **kw)

    def format(self, item):
        if item.audio:
            return icon('volume-up')

    def order(self):
        return Sentence_files.id

    def search(self, qs):
        if qs == 'yes':
            return Sentence_files.pk != 0


class Sentences(DataTable):

    """Default DataTable for Sentence objects."""

    __constraints__ = [Parameter, Language]

    def base_query(self, query):
        query = query\
            .outerjoin(
                Sentence_files,
                and_(
                    Sentence_files.object_pk == Sentence.pk,
                    Sentence_files.mime_type.contains('audio/')))\
            .options(joinedload(Sentence._files))

        if self.language:
            query = query.filter(Sentence.language_pk == self.language.pk)
        else:
            query = query.join(Language).options(joinedload(Sentence.language))

        if self.parameter:
            query = query.join(ValueSentence, Value, ValueSet)\
                .filter(ValueSet.parameter_pk == self.parameter.pk)

        return query

    def col_defs(self):
        return [
            Col(self, 'id', bSortable=False, input_size='mini'),
            LinkCol(self, 'name', sTitle='Primary text', sClass="object-language"),
            TsvCol(self, 'analyzed', sTitle='Analyzed text'),
            TsvCol(self, 'gloss', sClass="gloss"),
            Col(self,
                'description',
                sTitle=self.req.translate('Translation'),
                sClass="translation"),
            TypeCol(self, 'type'),
            LinkCol(
                self,
                'language',
                model_col=Language.name,
                get_obj=lambda i: i.language,
                bSortable=not self.language,
                bSearchable=not self.language),
            DetailsRowLinkCol(self, 'd'),
        ]

    def get_options(self):
        return {'aaSorting': []}



from sqlalchemy import Column, Float, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, validates, backref
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    unitparameter as _unitparameter)  # needed to initalize relationship
assert _unitparameter

__all__ = ('UnitValue',)


class UnitValue_data(Base, DataMixin):
    pass


class UnitValue_files(Base, FilesMixin):
    pass


@implementer(interfaces.IUnitValue)
class UnitValue(Base,
                PolymorphicBaseMixin,
                IdNameDescriptionMixin,
                HasDataMixin,
                HasFilesMixin):

    __table_args__ = (UniqueConstraint(
        'unit_pk', 'unitparameter_pk', 'contribution_pk', 'name', 'unitdomainelement_pk'),
    )

    unit_pk = Column(Integer, ForeignKey('unit.pk'), nullable=False)
    unitparameter_pk = Column(Integer, ForeignKey('unitparameter.pk'), nullable=False)
    contribution_pk = Column(Integer, ForeignKey('contribution.pk'))

    # Values may be taken from a domain.
    unitdomainelement_pk = Column(Integer, ForeignKey('unitdomainelement.pk'))

    # Languages may have multiple values for the same parameter. Their relative
    # frequency can be stored here.
    frequency = Column(Float)

    unitparameter = relationship('UnitParameter', innerjoin=True, backref='unitvalues')
    unitdomainelement = relationship('UnitDomainElement', backref='unitvalues')
    contribution = relationship('Contribution', backref='unitvalues')

    @declared_attr
    def unit(cls):
        return relationship(
            'Unit', innerjoin=True, backref=backref('unitvalues', order_by=cls.unit_pk))

    @validates('unitparameter_pk')
    def validate_parameter_pk(self, key, unitparameter_pk):
        """Validator to sync related parameter.

        We have to make sure, the parameter a value is tied to and the parameter a
        possible domainelement is tied to stay in sync.
        """
        if self.unitdomainelement and self.unitdomainelement.unitparameter_pk:
            assert self.unitdomainelement.unitparameter_pk == unitparameter_pk
        return unitparameter_pk

    def __str__(self):
        if self.unitdomainelement and self.unitdomainelement.name:
            return self.unitdomainelement.name
        else:
            return self.name or self.id


#
# TODO: UnitValueSentence!
#



import pathlib
import functools
import configparser


def get_config(p):
    """Read a config file.

    :return: dict of ('section.option', value) pairs.
    """
    cfg = {}
    parser = configparser.ConfigParser()
    parser.read_file(pathlib.Path(p).open(encoding='utf8'))

    for section in parser.sections():
        getters = {
            'int': functools.partial(parser.getint, section),
            'boolean': functools.partial(parser.getboolean, section),
            'float': functools.partial(parser.getfloat, section),
            'list': lambda option: parser.get(section, option).split(),
        }
        default = functools.partial(parser.get, section)
        for option in parser.options(section):
            type_ = option.rpartition('_')[2] if '_' in option else None
            value = getters.get(type_, default)(option)
            cfg['{0}.{1}'.format(section, option)] = value

    return cfg



"""
Functionality for the creation of bibliographical data using the Bibliographic Ontology.

.. seealso:: http://bibliontology.com/
"""
#: Map BibTeX entry types to the corresponding bibo types.
TYPE_MAP = {
    "book": "bibo:Book",
    "booklet": "bibo:Book",
    "misc": "bibo:Document",
    "article": "bibo:Article",
    "inbook": "bibo:Chapter",
    "manual": "bibo:Manual",
    "inproceedings": "bibo:Article",
    "conference": "bibo:Article",
    "unpublished": "bibo:Document",
    "masterthesis": "bibo:Thesis",
    "phdthesis": "bibo:Thesis",
    "proceedings": "bibo:Proceedings",
    "techreport": "bibo:Report",
    "incollection": "bibo:BookSection",
    "heading": "bibo:Collection",
    "subject": "skos:Concept",
    "person": "foaf:Person",
}


ADD_FIELD_MAP = {
    "unpublished": ("bibo:status", "http://purl.org/ontology/bibo/status/unpublished"),
    "masterthesis": ("bibo:degree", "http://purl.org/ontology/bibo/degrees/ma"),
    "phdthesis": ("bibo:degree", "http://purl.org/ontology/bibo/degrees/phd"),
}


FIELD_MAP = {
    "title": "dcterms:title",
    "author": ("dcterms:creator", ("foaf:Agent", "foaf:name")),
    "booktitle": "dcterms:title",
    "publisher": ("dcterms:publisher", ("foaf:Organization", "foaf:name")),
    "year": "dcterms:date",
    "month": "dcterms:date",
    "isbn": "bibo:isbn",
    "editor": ("bibo:editor", ("foaf:Person", "foaf:name")),
    "institution": ("dcterms:contributor", ("foaf:Organization", "foaf:name")),
    "volume": "bibo:volume",
    "type": "dcterms:type",
    "series": ("dcterms:isPartOf", ("bibo:Series", "dcterms:title")),
    "school": ("rdfs:seeAlso", ("foaf:Organization", "foaf:name")),
    "pages": "bibo:pages",
    "organization": ("bibo:organizer", ("foaf:Organization", "foaf:name")),
    "number": "bibo:number",
    "note": "skos:note",
    "journal": ("dcterms:isPartOf", ("bibo:Journal", "dcterms:title")),
    "edition": "bibo:edition",
    "chapter": "bibo:chapter",
    "address": "vcard:locality",
    "eprint": "rdfs:seeAlso",
    "crossref": "dcterms:isPartOf",
    "name": "foaf:name",
    "updated": "dcterms:modified",
    "bibliography": "foaf:page",
    "honor": "foaf:page",
    "biography": "foaf:page",
    "comment": "skos:note",
    "rev": "dcterms:modified",
    "date": "dcterms:date",
    "howpublished": "dcterms:publisher",
    "keywords": "dcterms:subject",
    "abstract": "bibo:abstract",
    "description": "dcterms:description",
}
